@@ -10,10 +10,11 @@ use {
   },
 };
 
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Default, PartialEq)]
 pub(crate) struct Config {
   pub(crate) alias_style: AliasStyle,
   pub(crate) allow_missing: bool,
+  pub(crate) ceiling: Option<PathBuf>,
   pub(crate) check: bool,
   pub(crate) color: Color,
   pub(crate) command_color: Option<ansi_term::Color>,
@@ -32,6 +33,7 @@ pub(crate) struct Config {
   pub(crate) no_aliases: bool,
   pub(crate) no_dependencies: bool,
   pub(crate) one: bool,
+  pub(crate) overrides: BTreeMap<String, String>,
   pub(crate) search_config: SearchConfig,
   pub(crate) shell: Option<String>,
   pub(crate) shell_args: Option<Vec<String>>,
@@ -62,6 +64,7 @@ mod cmd {
   pub(crate) const REQUEST: &str = "REQUEST";
   pub(crate) const SHOW: &str = "SHOW";
   pub(crate) const SUMMARY: &str = "SUMMARY";
+  pub(crate) const USAGE: &str = "USAGE";
   pub(crate) const VARIABLES: &str = "VARIABLES";
 
   pub(crate) const ALL: &[&str] = &[
@@ -92,6 +95,7 @@ mod arg {
   pub(crate) const ALIAS_STYLE: &str = "ALIAS_STYLE";
   pub(crate) const ALLOW_MISSING: &str = "ALLOW-MISSING";
   pub(crate) const ARGUMENTS: &str = "ARGUMENTS";
+  pub(crate) const CEILING: &str = "CEILING";
   pub(crate) const CHECK: &str = "CHECK";
   pub(crate) const CHOOSER: &str = "CHOOSER";
   pub(crate) const CLEAR_SHELL_ARGS: &str = "CLEAR-SHELL-ARGS";
@@ -161,6 +165,14 @@ impl Config {
           .help("Set list command alias display style")
           .conflicts_with(arg::NO_ALIASES),
       )
+      .arg(
+        Arg::new(arg::CEILING)
+          .long("ceiling")
+          .env("JUST_CEILING")
+          .action(ArgAction::Set)
+          .value_parser(value_parser!(PathBuf))
+          .help("Do not ascend above <CEILING> directory when searching for a justfile."),
+      )
       .arg(
         Arg::new(arg::CHECK)
           .long("check")
@@ -587,6 +599,16 @@ impl Config {
           .help("List names of available recipes")
           .help_heading(cmd::HEADING),
       )
+      .arg(
+        Arg::new(cmd::USAGE)
+          .long("usage")
+          .num_args(1..)
+          .value_name("PATH")
+          .action(ArgAction::Set)
+          .conflicts_with(arg::ARGUMENTS)
+          .help("Print recipe usage information")
+          .help_heading(cmd::HEADING),
+      )
       .arg(
         Arg::new(cmd::VARIABLES)
           .long("variables")
@@ -705,14 +727,12 @@ impl Config {
     } else if matches.get_flag(cmd::CHOOSE) {
       Subcommand::Choose {
         chooser: matches.get_one::<String>(arg::CHOOSER).map(Into::into),
-        overrides,
       }
     } else if let Some(values) = matches.get_many::<OsString>(cmd::COMMAND) {
       let mut arguments = values.map(Into::into).collect::<Vec<OsString>>();
       Subcommand::Command {
         binary: arguments.remove(0),
         arguments,
-        overrides,
       }
     } else if let Some(&shell) = matches.get_one::<completions::Shell>(cmd::COMPLETIONS) {
       Subcommand::Completions { shell }
@@ -734,7 +754,6 @@ impl Config {
 
       Subcommand::Evaluate {
         variable: positional.arguments.into_iter().next(),
-        overrides,
       }
     } else if matches.get_flag(cmd::FORMAT) {
       Subcommand::Format
@@ -759,12 +778,15 @@ impl Config {
       }
     } else if matches.get_flag(cmd::SUMMARY) {
       Subcommand::Summary
+    } else if let Some(path) = matches.get_many::<String>(cmd::USAGE) {
+      Subcommand::Usage {
+        path: Self::parse_module_path(path)?,
+      }
     } else if matches.get_flag(cmd::VARIABLES) {
       Subcommand::Variables
     } else {
       Subcommand::Run {
         arguments: positional.arguments,
-        overrides,
       }
     };
 
@@ -777,6 +799,7 @@ impl Config {
         .unwrap()
         .clone(),
       allow_missing: matches.get_flag(arg::ALLOW_MISSING),
+      ceiling: matches.get_one::<PathBuf>(arg::CEILING).cloned(),
       check: matches.get_flag(arg::CHECK),
       color: (*matches.get_one::<UseColor>(arg::COLOR).unwrap()).into(),
       command_color: matches
@@ -803,6 +826,7 @@ impl Config {
       no_aliases: matches.get_flag(arg::NO_ALIASES),
       no_dependencies: matches.get_flag(arg::NO_DEPS),
       one: matches.get_flag(arg::ONE),
+      overrides,
       search_config,
       shell: matches.get_one::<String>(arg::SHELL).map(Into::into),
       shell_args: if matches.get_flag(arg::CLEAR_SHELL_ARGS) {
@@ -861,6 +885,7 @@ mod tests {
       $(dump_format: $dump_format:expr,)?
       $(highlight: $highlight:expr,)?
       $(no_dependencies: $no_dependencies:expr,)?
+      $(overrides: $overrides:expr,)?
       $(search_config: $search_config:expr,)?
       $(shell: $shell:expr,)?
       $(shell_args: $shell_args:expr,)?
@@ -882,6 +907,7 @@ mod tests {
           $(dump_format: $dump_format,)?
           $(highlight: $highlight,)?
           $(no_dependencies: $no_dependencies,)?
+          $(overrides: $overrides,)?
           $(search_config: $search_config,)?
           $(shell: $shell,)?
           $(shell_args: $shell_args,)?
@@ -1139,45 +1165,45 @@ mod tests {
   test! {
     name: set_default,
     args: [],
+    overrides: map!(),
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!(),
     },
   }
 
   test! {
     name: set_one,
     args: ["--set", "foo", "bar"],
+    overrides: map!{"foo": "bar"},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{"foo": "bar"},
     },
   }
 
   test! {
     name: set_empty,
     args: ["--set", "foo", ""],
+    overrides: map!{"foo": ""},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{"foo": ""},
     },
   }
 
   test! {
     name: set_two,
     args: ["--set", "foo", "bar", "--set", "bar", "baz"],
+    overrides: map!{"foo": "bar", "bar": "baz"},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{"foo": "bar", "bar": "baz"},
     },
   }
 
   test! {
     name: set_override,
     args: ["--set", "foo", "bar", "--set", "foo", "baz"],
+    overrides: map!{"foo": "baz"},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{"foo": "baz"},
     },
   }
 
@@ -1239,9 +1265,9 @@ mod tests {
   test! {
     name: subcommand_default,
     args: [],
+    overrides: map!{},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{},
     },
   }
 
@@ -1333,8 +1359,8 @@ mod tests {
   test! {
     name: subcommand_evaluate,
     args: ["--evaluate"],
+    overrides: map!{},
     subcommand: Subcommand::Evaluate {
-      overrides: map!{},
       variable: None,
     },
   }
@@ -1342,8 +1368,8 @@ mod tests {
   test! {
     name: subcommand_evaluate_overrides,
     args: ["--evaluate", "x=y"],
+    overrides: map!{"x": "y"},
     subcommand: Subcommand::Evaluate {
-      overrides: map!{"x": "y"},
       variable: None,
     },
   }
@@ -1351,8 +1377,8 @@ mod tests {
   test! {
     name: subcommand_evaluate_overrides_with_argument,
     args: ["--evaluate", "x=y", "foo"],
+    overrides: map!{"x": "y"},
     subcommand: Subcommand::Evaluate {
-      overrides: map!{"x": "y"},
       variable: Some("foo".to_owned()),
     },
   }
@@ -1408,45 +1434,45 @@ mod tests {
   test! {
     name: arguments,
     args: ["foo", "bar"],
+    overrides: map!{},
     subcommand: Subcommand::Run {
       arguments: vec![String::from("foo"), String::from("bar")],
-      overrides: map!{},
     },
   }
 
   test! {
     name: arguments_leading_equals,
     args: ["=foo"],
+    overrides: map!{},
     subcommand: Subcommand::Run {
       arguments: vec!["=foo".to_owned()],
-      overrides: map!{},
     },
   }
 
   test! {
     name: overrides,
     args: ["foo=bar", "bar=baz"],
+    overrides: map!{"foo": "bar", "bar": "baz"},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{"foo": "bar", "bar": "baz"},
     },
   }
 
   test! {
     name: overrides_empty,
     args: ["foo=", "bar="],
+    overrides: map!{"foo": "", "bar": ""},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{"foo": "", "bar": ""},
     },
   }
 
   test! {
     name: overrides_override_sets,
     args: ["--set", "foo", "0", "--set", "bar", "1", "foo=bar", "bar=baz"],
+    overrides: map!{"foo": "bar", "bar": "baz"},
     subcommand: Subcommand::Run {
       arguments: Vec::new(),
-      overrides: map!{"foo": "bar", "bar": "baz"},
     },
   }
 
@@ -1547,7 +1573,7 @@ mod tests {
     search_config: SearchConfig::FromSearchDirectory {
       search_directory: PathBuf::from(".."),
     },
-    subcommand: Subcommand::Run { arguments: vec!["build".to_owned()], overrides: BTreeMap::new() },
+    subcommand: Subcommand::Run { arguments: vec!["build".to_owned()] },
   }
 
   test! {
@@ -1572,7 +1598,7 @@ mod tests {
     search_config: SearchConfig::FromSearchDirectory {
       search_directory: PathBuf::from("foo"),
     },
-    subcommand: Subcommand::Run { arguments: vec!["build".to_owned()], overrides: BTreeMap::new() },
+    subcommand: Subcommand::Run { arguments: vec!["build".to_owned()] },
   }
 
   error! {
