@@ -1,7 +1,4 @@
-use crate::common::*;
-
-use CompilationErrorKind::*;
-use TokenKind::*;
+use {super::*, CompileErrorKind::*, TokenKind::*};
 
 /// Just language lexer
 ///
@@ -12,48 +9,59 @@ use TokenKind::*;
 /// slight against regular expressions, the lexer was just idiosyncratically
 /// bad.
 pub(crate) struct Lexer<'src> {
-  /// Source text
-  src:                 &'src str,
   /// Char iterator
-  chars:               Chars<'src>,
-  /// Tokens
-  tokens:              Vec<Token<'src>>,
-  /// Current token start
-  token_start:         Position,
-  /// Current token end
-  token_end:           Position,
-  /// Next character to be lexed
-  next:                Option<char>,
-  /// Next indent will start a recipe body
-  recipe_body_pending: bool,
-  /// Inside recipe body
-  recipe_body:         bool,
+  chars: Chars<'src>,
   /// Indentation stack
-  indentation:         Vec<&'src str>,
+  indentation: Vec<&'src str>,
   /// Interpolation token start stack
   interpolation_stack: Vec<Token<'src>>,
+  /// Next character to be lexed
+  next: Option<char>,
   /// Current open delimiters
-  open_delimiters:     Vec<(Delimiter, usize)>,
+  open_delimiters: Vec<(Delimiter, usize)>,
+  /// Path to source file
+  path: &'src Path,
+  /// Inside recipe body
+  recipe_body: bool,
+  /// Next indent will start a recipe body
+  recipe_body_pending: bool,
+  /// Source text
+  src: &'src str,
+  /// Current token end
+  token_end: Position,
+  /// Current token start
+  token_start: Position,
+  /// Tokens
+  tokens: Vec<Token<'src>>,
 }
 
 impl<'src> Lexer<'src> {
-  /// Lex `text`
-  pub(crate) fn lex(src: &str) -> CompilationResult<Vec<Token>> {
-    Lexer::new(src).tokenize()
+  pub(crate) const INTERPOLATION_END: &'static str = "}}";
+  pub(crate) const INTERPOLATION_ESCAPE: &'static str = "{{{{";
+  pub(crate) const INTERPOLATION_START: &'static str = "{{";
+
+  /// Lex `src`
+  pub(crate) fn lex(path: &'src Path, src: &'src str) -> CompileResult<'src, Vec<Token<'src>>> {
+    Self::new(path, src).tokenize()
   }
 
-  /// Create a new Lexer to lex `text`
-  fn new(src: &'src str) -> Lexer<'src> {
+  #[cfg(test)]
+  pub(crate) fn test_lex(src: &'src str) -> CompileResult<'src, Vec<Token<'src>>> {
+    Self::new("justfile".as_ref(), src).tokenize()
+  }
+
+  /// Create a new Lexer to lex `src`
+  fn new(path: &'src Path, src: &'src str) -> Self {
     let mut chars = src.chars();
     let next = chars.next();
 
     let start = Position {
       offset: 0,
       column: 0,
-      line:   0,
+      line: 0,
     };
 
-    Lexer {
+    Self {
       indentation: vec![""],
       tokens: Vec::new(),
       token_start: start,
@@ -65,12 +73,13 @@ impl<'src> Lexer<'src> {
       chars,
       next,
       src,
+      path,
     }
   }
 
   /// Advance over the character in `self.next`, updating `self.token_end`
   /// accordingly.
-  fn advance(&mut self) -> CompilationResult<'src, ()> {
+  fn advance(&mut self) -> CompileResult<'src> {
     match self.next {
       Some(c) => {
         let len_utf8 = c.len_utf8();
@@ -86,20 +95,11 @@ impl<'src> Lexer<'src> {
         self.next = self.chars.next();
 
         Ok(())
-      },
+      }
       None => Err(self.internal_error("Lexer advanced past end of text")),
     }
   }
 
-  /// Advance over N characters.
-  fn skip(&mut self, n: usize) -> CompilationResult<'src, ()> {
-    for _ in 0..n {
-      self.advance()?;
-    }
-
-    Ok(())
-  }
-
   /// Lexeme of in-progress token
   fn lexeme(&self) -> &'src str {
     &self.src[self.token_start.offset..self.token_end.offset]
@@ -110,7 +110,7 @@ impl<'src> Lexer<'src> {
     self.token_end.offset - self.token_start.offset
   }
 
-  fn accepted(&mut self, c: char) -> CompilationResult<'src, bool> {
+  fn accepted(&mut self, c: char) -> CompileResult<'src, bool> {
     if self.next_is(c) {
       self.advance()?;
       Ok(true)
@@ -119,9 +119,9 @@ impl<'src> Lexer<'src> {
     }
   }
 
-  fn presume(&mut self, c: char) -> CompilationResult<'src, ()> {
+  fn presume(&mut self, c: char) -> CompileResult<'src> {
     if !self.next_is(c) {
-      return Err(self.internal_error(format!("Lexer presumed character `{}`", c)));
+      return Err(self.internal_error(format!("Lexer presumed character `{c}`")));
     }
 
     self.advance()?;
@@ -129,7 +129,7 @@ impl<'src> Lexer<'src> {
     Ok(())
   }
 
-  fn presume_str(&mut self, s: &str) -> CompilationResult<'src, ()> {
+  fn presume_str(&mut self, s: &str) -> CompileResult<'src> {
     for c in s.chars() {
       self.presume(c)?;
     }
@@ -162,14 +162,19 @@ impl<'src> Lexer<'src> {
     self.next_is('\n') || self.rest_starts_with("\r\n")
   }
 
+  /// Are we at end-of-file?
+  fn at_eof(&self) -> bool {
+    self.rest().is_empty()
+  }
+
   /// Are we at end-of-line or end-of-file?
   fn at_eol_or_eof(&self) -> bool {
-    self.at_eol() || self.rest().is_empty()
+    self.at_eol() || self.at_eof()
   }
 
   /// Get current indentation
   fn indentation(&self) -> &'src str {
-    self.indentation.last().cloned().unwrap()
+    self.indentation.last().unwrap()
   }
 
   /// Are we currently indented
@@ -181,12 +186,13 @@ impl<'src> Lexer<'src> {
   /// and `self.token_end`
   fn token(&mut self, kind: TokenKind) {
     self.tokens.push(Token {
-      offset: self.token_start.offset,
       column: self.token_start.column,
-      line: self.token_start.line,
-      src: self.src,
-      length: self.token_end.offset - self.token_start.offset,
       kind,
+      length: self.token_end.offset - self.token_start.offset,
+      line: self.token_start.line,
+      offset: self.token_start.offset,
+      path: self.path,
+      src: self.src,
     });
 
     // Set `token_start` to point after the lexed token
@@ -194,38 +200,37 @@ impl<'src> Lexer<'src> {
   }
 
   /// Create an internal error with `message`
-  fn internal_error(&self, message: impl Into<String>) -> CompilationError<'src> {
+  fn internal_error(&self, message: impl Into<String>) -> CompileError<'src> {
     // Use `self.token_end` as the location of the error
     let token = Token {
-      src:    self.src,
+      src: self.src,
       offset: self.token_end.offset,
-      line:   self.token_end.line,
+      line: self.token_end.line,
       column: self.token_end.column,
       length: 0,
-      kind:   Unspecified,
+      kind: Unspecified,
+      path: self.path,
     };
-    CompilationError {
-      kind: CompilationErrorKind::Internal {
+    CompileError::new(
+      token,
+      Internal {
         message: message.into(),
       },
-      token,
-    }
+    )
   }
 
   /// Create a compilation error with `kind`
-  fn error(&self, kind: CompilationErrorKind<'src>) -> CompilationError<'src> {
+  fn error(&self, kind: CompileErrorKind<'src>) -> CompileError<'src> {
     // Use the in-progress token span as the location of the error.
 
     // The width of the error site to highlight depends on the kind of error:
     let length = match kind {
       UnterminatedString | UnterminatedBacktick => {
-        let kind = match StringKind::from_token_start(self.lexeme()) {
-          Some(kind) => kind,
-          None =>
-            return self.internal_error("Lexer::error: expected string or backtick token start"),
+        let Some(kind) = StringKind::from_token_start(self.lexeme()) else {
+          return self.internal_error("Lexer::error: expected string or backtick token start");
         };
         kind.delimiter().len()
-      },
+      }
       // highlight the full token
       _ => self.lexeme().len(),
     };
@@ -237,26 +242,19 @@ impl<'src> Lexer<'src> {
       line: self.token_start.line,
       column: self.token_start.column,
       length,
+      path: self.path,
     };
 
-    CompilationError { token, kind }
+    CompileError::new(token, kind)
   }
 
-  fn unterminated_interpolation_error(interpolation_start: Token<'src>) -> CompilationError<'src> {
-    CompilationError {
-      token: interpolation_start,
-      kind:  UnterminatedInterpolation,
-    }
+  fn unterminated_interpolation_error(interpolation_start: Token<'src>) -> CompileError<'src> {
+    CompileError::new(interpolation_start, UnterminatedInterpolation)
   }
 
   /// True if `text` could be an identifier
   pub(crate) fn is_identifier(text: &str) -> bool {
-    if !text
-      .chars()
-      .next()
-      .map(Self::is_identifier_start)
-      .unwrap_or(false)
-    {
+    if !text.chars().next().is_some_and(Self::is_identifier_start) {
       return false;
     }
 
@@ -270,21 +268,17 @@ impl<'src> Lexer<'src> {
   }
 
   /// True if `c` can be the first character of an identifier
-  fn is_identifier_start(c: char) -> bool {
+  pub(crate) fn is_identifier_start(c: char) -> bool {
     matches!(c, 'a'..='z' | 'A'..='Z' | '_')
   }
 
-  /// True if `c` can be a continuation character of an idenitifier
-  fn is_identifier_continue(c: char) -> bool {
-    if Self::is_identifier_start(c) {
-      return true;
-    }
-
-    matches!(c, '0'..='9' | '-')
+  /// True if `c` can be a continuation character of an identifier
+  pub(crate) fn is_identifier_continue(c: char) -> bool {
+    Self::is_identifier_start(c) || matches!(c, '0'..='9' | '-')
   }
 
   /// Consume the text and produce a series of tokens
-  fn tokenize(mut self) -> CompilationResult<'src, Vec<Token<'src>>> {
+  fn tokenize(mut self) -> CompileResult<'src, Vec<Token<'src>>> {
     loop {
       if self.token_start.column == 0 {
         self.lex_line_start()?;
@@ -293,13 +287,13 @@ impl<'src> Lexer<'src> {
       match self.next {
         Some(first) => {
           if let Some(&interpolation_start) = self.interpolation_stack.last() {
-            self.lex_interpolation(interpolation_start, first)?
+            self.lex_interpolation(interpolation_start, first)?;
           } else if self.recipe_body {
-            self.lex_body()?
+            self.lex_body()?;
           } else {
-            self.lex_normal(first)?
-          };
-        },
+            self.lex_normal(first)?;
+          }
+        }
         None => break,
       }
     }
@@ -322,7 +316,7 @@ impl<'src> Lexer<'src> {
   }
 
   /// Handle blank lines and indentation
-  fn lex_line_start(&mut self) -> CompilationResult<'src, ()> {
+  fn lex_line_start(&mut self) -> CompileResult<'src> {
     enum Indentation<'src> {
       // Line only contains whitespace
       Blank,
@@ -352,6 +346,18 @@ impl<'src> Lexer<'src> {
 
     let whitespace = &self.rest()[..nonblank_index];
 
+    if self.open_delimiters_or_interpolation() {
+      if !whitespace.is_empty() {
+        while self.next_is_whitespace() {
+          self.advance()?;
+        }
+
+        self.token(Whitespace);
+      }
+
+      return Ok(());
+    }
+
     let body_whitespace = &whitespace[..whitespace
       .char_indices()
       .take(self.indentation().chars().count())
@@ -403,10 +409,10 @@ impl<'src> Lexer<'src> {
           }
 
           self.token(Whitespace);
-        };
+        }
 
         Ok(())
-      },
+      }
       Continue => {
         if !self.indentation().is_empty() {
           for _ in self.indentation().chars() {
@@ -417,7 +423,7 @@ impl<'src> Lexer<'src> {
         }
 
         Ok(())
-      },
+      }
       Decrease => {
         while self.indentation() != whitespace {
           self.lex_dedent();
@@ -432,14 +438,14 @@ impl<'src> Lexer<'src> {
         }
 
         Ok(())
-      },
+      }
       Mixed { whitespace } => {
         for _ in whitespace.chars() {
           self.advance()?;
         }
 
         Err(self.error(MixedLeadingWhitespace { whitespace }))
-      },
+      }
       Inconsistent => {
         for _ in whitespace.chars() {
           self.advance()?;
@@ -447,58 +453,79 @@ impl<'src> Lexer<'src> {
 
         Err(self.error(InconsistentLeadingWhitespace {
           expected: self.indentation(),
-          found:    whitespace,
+          found: whitespace,
         }))
-      },
+      }
       Increase => {
         while self.next_is_whitespace() {
           self.advance()?;
         }
 
-        if self.open_delimiters() {
-          self.token(Whitespace);
-        } else {
-          let indentation = self.lexeme();
-          self.indentation.push(indentation);
-          self.token(Indent);
-          if self.recipe_body_pending {
-            self.recipe_body = true;
-          }
+        let indentation = self.lexeme();
+        self.indentation.push(indentation);
+        self.token(Indent);
+        if self.recipe_body_pending {
+          self.recipe_body = true;
         }
 
         Ok(())
-      },
+      }
     }
   }
 
   /// Lex token beginning with `start` outside of a recipe body
-  fn lex_normal(&mut self, start: char) -> CompilationResult<'src, ()> {
+  fn lex_normal(&mut self, start: char) -> CompileResult<'src> {
     match start {
-      '!' => self.lex_bang(),
-      '*' => self.lex_single(Asterisk),
+      ' ' | '\t' => self.lex_whitespace(),
+      '!' if self.rest().starts_with("!include") => Err(self.error(Include)),
+      '!' => self.lex_choices('!', &[('=', BangEquals), ('~', BangTilde)], None),
+      '#' => self.lex_comment(),
       '$' => self.lex_single(Dollar),
-      '@' => self.lex_single(At),
-      '[' => self.lex_delimiter(BracketL),
-      ']' => self.lex_delimiter(BracketR),
-      '=' => self.lex_choice('=', EqualsEquals, Equals),
-      ',' => self.lex_single(Comma),
-      ':' => self.lex_colon(),
+      '&' => self.lex_digraph('&', '&', AmpersandAmpersand),
       '(' => self.lex_delimiter(ParenL),
       ')' => self.lex_delimiter(ParenR),
-      '{' => self.lex_delimiter(BraceL),
-      '}' => self.lex_delimiter(BraceR),
+      '*' => self.lex_single(Asterisk),
       '+' => self.lex_single(Plus),
-      '#' => self.lex_comment(),
-      ' ' => self.lex_whitespace(),
-      '`' | '"' | '\'' => self.lex_string(),
-      '\n' => self.lex_eol(),
-      '\r' => self.lex_eol(),
-      '\t' => self.lex_whitespace(),
+      ',' => self.lex_single(Comma),
+      '/' => self.lex_single(Slash),
+      ':' => self.lex_colon(),
+      '=' => self.lex_choices(
+        '=',
+        &[('=', EqualsEquals), ('~', EqualsTilde)],
+        Some(Equals),
+      ),
+      '?' => self.lex_single(QuestionMark),
+      '@' => self.lex_single(At),
+      '[' => self.lex_delimiter(BracketL),
+      '\\' => self.lex_escape(),
+      '\n' | '\r' => self.lex_eol(),
+      '\u{feff}' => self.lex_single(ByteOrderMark),
+      ']' => self.lex_delimiter(BracketR),
+      '`' | '"' | '\'' => self.lex_string(None),
+      '{' => self.lex_delimiter(BraceL),
+      '|' => self.lex_digraph('|', '|', BarBar),
+      '}' => {
+        let format_string_kind = self.open_delimiters.last().and_then(|(delimiter, _line)| {
+          if !self.rest().starts_with(Self::INTERPOLATION_END) {
+            None
+          } else if let Delimiter::FormatString(kind) = delimiter {
+            Some(kind)
+          } else {
+            None
+          }
+        });
+
+        if let Some(format_string_kind) = format_string_kind {
+          self.lex_string(Some(*format_string_kind))
+        } else {
+          self.lex_delimiter(BraceR)
+        }
+      }
       _ if Self::is_identifier_start(start) => self.lex_identifier(),
       _ => {
         self.advance()?;
-        Err(self.error(UnknownStartOfToken))
-      },
+        Err(self.error(UnknownStartOfToken { start }))
+      }
     }
   }
 
@@ -507,19 +534,18 @@ impl<'src> Lexer<'src> {
     &mut self,
     interpolation_start: Token<'src>,
     start: char,
-  ) -> CompilationResult<'src, ()> {
-    if self.rest_starts_with("}}") {
+  ) -> CompileResult<'src> {
+    if self.rest_starts_with(Self::INTERPOLATION_END) && self.open_delimiters.is_empty() {
       // end current interpolation
       if self.interpolation_stack.pop().is_none() {
-        self.advance()?;
-        self.advance()?;
+        self.presume_str(Self::INTERPOLATION_END)?;
         return Err(self.internal_error(
           "Lexer::lex_interpolation found `}}` but was called with empty interpolation stack.",
         ));
       }
       // Emit interpolation end token
       self.lex_double(InterpolationEnd)
-    } else if self.at_eol_or_eof() {
+    } else if self.at_eof() && self.open_delimiters.is_empty() {
       // Return unterminated interpolation error that highlights the opening
       // {{
       Err(Self::unterminated_interpolation_error(interpolation_start))
@@ -530,19 +556,19 @@ impl<'src> Lexer<'src> {
   }
 
   /// Lex token while in recipe body
-  fn lex_body(&mut self) -> CompilationResult<'src, ()> {
+  fn lex_body(&mut self) -> CompileResult<'src> {
     enum Terminator {
+      EndOfFile,
+      Interpolation,
       Newline,
       NewlineCarriageReturn,
-      Interpolation,
-      EndOfFile,
     }
 
     use Terminator::*;
 
     let terminator = loop {
-      if self.rest_starts_with("{{{{") {
-        self.skip(4)?;
+      if self.rest_starts_with(Self::INTERPOLATION_ESCAPE) {
+        self.presume_str(Self::INTERPOLATION_ESCAPE)?;
         continue;
       }
 
@@ -554,11 +580,11 @@ impl<'src> Lexer<'src> {
         break NewlineCarriageReturn;
       }
 
-      if self.rest_starts_with("{{") {
+      if self.rest_starts_with(Self::INTERPOLATION_START) {
         break Interpolation;
       }
 
-      if self.rest().is_empty() {
+      if self.at_eof() {
         break EndOfFile;
       }
 
@@ -579,7 +605,7 @@ impl<'src> Lexer<'src> {
           .interpolation_stack
           .push(self.tokens[self.tokens.len() - 1]);
         Ok(())
-      },
+      }
       EndOfFile => Ok(()),
     }
   }
@@ -593,14 +619,14 @@ impl<'src> Lexer<'src> {
   }
 
   /// Lex a single-character token
-  fn lex_single(&mut self, kind: TokenKind) -> CompilationResult<'src, ()> {
+  fn lex_single(&mut self, kind: TokenKind) -> CompileResult<'src> {
     self.advance()?;
     self.token(kind);
     Ok(())
   }
 
   /// Lex a double-character token
-  fn lex_double(&mut self, kind: TokenKind) -> CompilationResult<'src, ()> {
+  fn lex_double(&mut self, kind: TokenKind) -> CompileResult<'src> {
     self.advance()?;
     self.advance()?;
     self.token(kind);
@@ -610,43 +636,63 @@ impl<'src> Lexer<'src> {
   /// Lex a double-character token of kind `then` if the second character of
   /// that token would be `second`, otherwise lex a single-character token of
   /// kind `otherwise`
-  fn lex_choice(
+  fn lex_choices(
     &mut self,
-    second: char,
-    then: TokenKind,
-    otherwise: TokenKind,
-  ) -> CompilationResult<'src, ()> {
-    self.advance()?;
+    first: char,
+    choices: &[(char, TokenKind)],
+    otherwise: Option<TokenKind>,
+  ) -> CompileResult<'src> {
+    self.presume(first)?;
 
-    if self.accepted(second)? {
-      self.token(then);
+    for (second, then) in choices {
+      if self.accepted(*second)? {
+        self.token(*then);
+        return Ok(());
+      }
+    }
+
+    if let Some(token) = otherwise {
+      self.token(token);
     } else {
-      self.token(otherwise);
+      // Emit an unspecified token to consume the current character,
+      self.token(Unspecified);
+
+      let expected = choices.iter().map(|choice| choice.0).collect();
+
+      if self.at_eof() {
+        return Err(self.error(UnexpectedEndOfToken { expected }));
+      }
+
+      // …and advance past another character,
+      self.advance()?;
+
+      // …so that the error we produce highlights the unexpected character.
+      return Err(self.error(UnexpectedCharacter { expected }));
     }
 
     Ok(())
   }
 
   /// Lex an opening or closing delimiter
-  fn lex_delimiter(&mut self, kind: TokenKind) -> CompilationResult<'src, ()> {
-    use Delimiter::*;
-
+  fn lex_delimiter(&mut self, kind: TokenKind) -> CompileResult<'src> {
     match kind {
-      BraceL => self.open_delimiter(Brace),
-      BraceR => self.close_delimiter(Brace)?,
-      BracketL => self.open_delimiter(Bracket),
-      BracketR => self.close_delimiter(Bracket)?,
-      ParenL => self.open_delimiter(Paren),
-      ParenR => self.close_delimiter(Paren)?,
-      _ =>
+      BraceL => self.open_delimiter(Delimiter::Brace),
+      BraceR => self.close_delimiter(Delimiter::Brace)?,
+      BracketL => self.open_delimiter(Delimiter::Bracket),
+      BracketR => self.close_delimiter(Delimiter::Bracket)?,
+      ParenL => self.open_delimiter(Delimiter::Paren),
+      ParenR => self.close_delimiter(Delimiter::Paren)?,
+      _ => {
         return Err(self.internal_error(format!(
-          "Lexer::lex_delimiter called with non-delimiter token: `{}`",
-          kind,
-        ))),
+          "Lexer::lex_delimiter called with non-delimiter token: `{kind}`",
+        )));
+      }
     }
 
     // Emit the delimiter token
-    self.lex_single(kind)
+    self.lex_single(kind)?;
+
+    Ok(())
   }
 
   /// Push a delimiter onto the open delimiter stack
@@ -657,7 +703,7 @@ impl<'src> Lexer<'src> {
   }
 
   /// Pop a delimiter from the open delimiter stack and error if incorrect type
-  fn close_delimiter(&mut self, close: Delimiter) -> CompilationResult<'src, ()> {
+  fn close_delimiter(&mut self, close: Delimiter) -> CompileResult<'src> {
     match self.open_delimiters.pop() {
       Some((open, _)) if open == close => Ok(()),
       Some((open, open_line)) => Err(self.error(MismatchedClosingDelimiter {
@@ -670,33 +716,45 @@ impl<'src> Lexer<'src> {
   }
 
   /// Return true if there are any unclosed delimiters
-  fn open_delimiters(&self) -> bool {
-    !self.open_delimiters.is_empty()
+  fn open_delimiters_or_interpolation(&self) -> bool {
+    !self.open_delimiters.is_empty() || !self.interpolation_stack.is_empty()
   }
 
-  /// Lex a token starting with '!'
-  fn lex_bang(&mut self) -> CompilationResult<'src, ()> {
-    self.presume('!')?;
+  /// Lex a two-character digraph
+  fn lex_digraph(&mut self, left: char, right: char, token: TokenKind) -> CompileResult<'src> {
+    self.presume(left)?;
 
-    if self.accepted('=')? {
-      self.token(BangEquals);
+    if self.accepted(right)? {
+      self.token(token);
       Ok(())
     } else {
       // Emit an unspecified token to consume the current character,
       self.token(Unspecified);
+
+      if self.at_eof() {
+        return Err(self.error(UnexpectedEndOfToken {
+          expected: vec![right],
+        }));
+      }
+
       // …and advance past another character,
       self.advance()?;
+
       // …so that the error we produce highlights the unexpected character.
-      Err(self.error(UnexpectedCharacter { expected: '=' }))
+      Err(self.error(UnexpectedCharacter {
+        expected: vec![right],
+      }))
     }
   }
 
   /// Lex a token starting with ':'
-  fn lex_colon(&mut self) -> CompilationResult<'src, ()> {
+  fn lex_colon(&mut self) -> CompileResult<'src> {
     self.presume(':')?;
 
     if self.accepted('=')? {
       self.token(ColonEquals);
+    } else if self.accepted(':')? {
+      self.token(ColonColon);
     } else {
       self.token(Colon);
       self.recipe_body_pending = true;
@@ -705,8 +763,33 @@ impl<'src> Lexer<'src> {
     Ok(())
   }
 
+  /// Lex an token starting with '\' escape
+  fn lex_escape(&mut self) -> CompileResult<'src> {
+    self.presume('\\')?;
+
+    // Treat newline escaped with \ as whitespace
+    if self.accepted('\n')? {
+      while self.next_is_whitespace() {
+        self.advance()?;
+      }
+      self.token(Whitespace);
+    } else if self.accepted('\r')? {
+      if !self.accepted('\n')? {
+        return Err(self.error(UnpairedCarriageReturn));
+      }
+      while self.next_is_whitespace() {
+        self.advance()?;
+      }
+      self.token(Whitespace);
+    } else if let Some(character) = self.next {
+      return Err(self.error(InvalidEscapeSequence { character }));
+    }
+
+    Ok(())
+  }
+
   /// Lex a carriage return and line feed
-  fn lex_eol(&mut self) -> CompilationResult<'src, ()> {
+  fn lex_eol(&mut self) -> CompileResult<'src> {
     if self.accepted('\r')? {
       if !self.accepted('\n')? {
         return Err(self.error(UnpairedCarriageReturn));
@@ -715,9 +798,8 @@ impl<'src> Lexer<'src> {
       self.presume('\n')?;
     }
 
-    // Emit an eol if there are no open delimiters, otherwise emit a whitespace
-    // token.
-    if self.open_delimiters() {
+    // Emit eol if there are no open delimiters, otherwise emit whitespace.
+    if self.open_delimiters_or_interpolation() {
       self.token(Whitespace);
     } else {
       self.token(Eol);
@@ -727,7 +809,7 @@ impl<'src> Lexer<'src> {
   }
 
   /// Lex name: [a-zA-Z_][a-zA-Z0-9_]*
-  fn lex_identifier(&mut self) -> CompilationResult<'src, ()> {
+  fn lex_identifier(&mut self) -> CompileResult<'src> {
     self.advance()?;
 
     while let Some(c) = self.next {
@@ -744,7 +826,7 @@ impl<'src> Lexer<'src> {
   }
 
   /// Lex comment: #[^\r\n]
-  fn lex_comment(&mut self) -> CompilationResult<'src, ()> {
+  fn lex_comment(&mut self) -> CompileResult<'src> {
     self.presume('#')?;
 
     while !self.at_eol_or_eof() {
@@ -757,9 +839,9 @@ impl<'src> Lexer<'src> {
   }
 
   /// Lex whitespace: [ \t]+
-  fn lex_whitespace(&mut self) -> CompilationResult<'src, ()> {
+  fn lex_whitespace(&mut self) -> CompileResult<'src> {
     while self.next_is_whitespace() {
-      self.advance()?
+      self.advance()?;
     }
 
     self.token(Whitespace);
@@ -769,27 +851,45 @@ impl<'src> Lexer<'src> {
 
   /// Lex a backtick, cooked string, or raw string.
   ///
-  /// Backtick:      `[^`]*`
+  /// Backtick:      ``[^`]*``
   /// Cooked string: "[^"]*" # also processes escape sequences
   /// Raw string:    '[^']*'
-  fn lex_string(&mut self) -> CompilationResult<'src, ()> {
-    let kind = if let Some(kind) = StringKind::from_token_start(self.rest()) {
+  fn lex_string(&mut self, format_string_kind: Option<StringKind>) -> CompileResult<'src> {
+    let format = format_string_kind.is_some()
+      || self.tokens.last().is_some_and(|token| {
+        token.kind == TokenKind::Identifier && token.lexeme() == Keyword::F.lexeme()
+      });
+
+    let kind = if let Some(kind) = format_string_kind {
+      self.presume_str(Self::INTERPOLATION_END)?;
       kind
     } else {
-      self.advance()?;
-      return Err(self.internal_error("Lexer::lex_string: invalid string start"));
+      let Some(kind) = StringKind::from_token_start(self.rest()) else {
+        self.advance()?;
+        return Err(self.internal_error("Lexer::lex_string: invalid string start"));
+      };
+      self.presume_str(kind.delimiter())?;
+      kind
     };
 
-    self.presume_str(kind.delimiter())?;
-
     let mut escape = false;
 
     loop {
-      if self.next == None {
+      if self.next.is_none() {
         return Err(self.error(kind.unterminated_error_kind()));
-      } else if kind.processes_escape_sequences() && self.next_is('\\') && !escape {
+      } else if !escape && kind.processes_escape_sequences() && self.next_is('\\') {
         escape = true;
-      } else if self.rest_starts_with(kind.delimiter()) && !escape {
+      } else if escape && kind.processes_escape_sequences() && self.next_is('u') {
+        escape = false;
+      } else if format && self.rest_starts_with(Self::INTERPOLATION_ESCAPE) {
+        escape = false;
+        self.advance()?;
+        self.advance()?;
+        self.advance()?;
+      } else if !escape
+        && (self.rest_starts_with(kind.delimiter())
+          || format && self.rest_starts_with(Self::INTERPOLATION_START))
+      {
         break;
       } else {
         escape = false;
@@ -798,8 +898,24 @@ impl<'src> Lexer<'src> {
       self.advance()?;
     }
 
-    self.presume_str(kind.delimiter())?;
-    self.token(kind.token_kind());
+    if format && self.rest_starts_with(Self::INTERPOLATION_START) {
+      self.presume_str(Self::INTERPOLATION_START)?;
+      if format_string_kind.is_some() {
+        self.token(FormatStringContinue);
+      } else {
+        self.token(FormatStringStart);
+        self.open_delimiter(Delimiter::FormatString(kind));
+      }
+    } else {
+      self.presume_str(kind.delimiter())?;
+
+      if let Some(format_string_kind) = format_string_kind {
+        self.close_delimiter(Delimiter::FormatString(format_string_kind))?;
+        self.token(FormatStringEnd);
+      } else {
+        self.token(kind.token_kind());
+      }
+    }
 
     Ok(())
   }
@@ -856,6 +972,7 @@ mod tests {
     }
   }
 
+  #[track_caller]
   fn test(text: &str, unindent_text: bool, want_kinds: &[TokenKind], want_lexemes: &[&str]) {
     let text = if unindent_text {
       unindent(text)
@@ -863,17 +980,14 @@ mod tests {
       text.to_owned()
     };
 
-    let have = Lexer::lex(&text).unwrap();
+    let have = Lexer::test_lex(&text).unwrap();
 
     let have_kinds = have
       .iter()
       .map(|token| token.kind)
       .collect::<Vec<TokenKind>>();
 
-    let have_lexemes = have
-      .iter()
-      .map(|token| token.lexeme())
-      .collect::<Vec<&str>>();
+    let have_lexemes = have.iter().map(Token::lexeme).collect::<Vec<&str>>();
 
     assert_eq!(have_kinds, want_kinds, "Token kind mismatch");
     assert_eq!(have_lexemes, want_lexemes, "Token lexeme mismatch");
@@ -912,34 +1026,44 @@ mod tests {
   fn default_lexeme(kind: TokenKind) -> &'static str {
     match kind {
       // Fixed lexemes
+      AmpersandAmpersand => "&&",
       Asterisk => "*",
       At => "@",
       BangEquals => "!=",
+      BangTilde => "!~",
+      BarBar => "||",
       BraceL => "{",
       BraceR => "}",
       BracketL => "[",
       BracketR => "]",
+      ByteOrderMark => "\u{feff}",
       Colon => ":",
+      ColonColon => "::",
       ColonEquals => ":=",
       Comma => ",",
       Dollar => "$",
       Eol => "\n",
       Equals => "=",
       EqualsEquals => "==",
+      EqualsTilde => "=~",
       Indent => "  ",
       InterpolationEnd => "}}",
       InterpolationStart => "{{",
       ParenL => "(",
       ParenR => ")",
       Plus => "+",
+      QuestionMark => "?",
+      Slash => "/",
       Whitespace => " ",
 
       // Empty lexemes
       Dedent | Eof => "",
 
       // Variable lexemes
-      Text | StringToken | Backtick | Identifier | Comment | Unspecified =>
-        panic!("Token {:?} has no default lexeme", kind),
+      Backtick | Comment | FormatStringContinue | FormatStringEnd | FormatStringStart
+      | Identifier | StringToken | Text | Unspecified => {
+        panic!("Token {kind:?} has no default lexeme")
+      }
     }
   }
 
@@ -960,18 +1084,19 @@ mod tests {
     };
   }
 
+  #[track_caller]
   fn error(
     src: &str,
     offset: usize,
     line: usize,
     column: usize,
     length: usize,
-    kind: CompilationErrorKind,
+    kind: CompileErrorKind,
   ) {
-    match Lexer::lex(src) {
+    match Lexer::test_lex(src) {
       Ok(_) => panic!("Lexing succeeded but expected"),
       Err(have) => {
-        let want = CompilationError {
+        let want = CompileError {
           token: Token {
             kind: have.token.kind,
             src,
@@ -979,11 +1104,12 @@ mod tests {
             line,
             column,
             length,
+            path: "justfile".as_ref(),
           },
-          kind,
+          kind: kind.into(),
         };
         assert_eq!(have, want);
-      },
+      }
     }
   }
 
@@ -1041,6 +1167,12 @@ mod tests {
     tokens: (StringToken:"\"\"\"hello\ngoodbye\"\"\""),
   }
 
+  test! {
+    name:   ampersand_ampersand,
+    text:   "&&",
+    tokens: (AmpersandAmpersand),
+  }
+
   test! {
     name:   equals,
     text:   "=",
@@ -1090,7 +1222,7 @@ mod tests {
   }
 
   test! {
-    name:   export_concatination,
+    name:   export_concatenation,
     text:   "export foo = 'foo' + 'bar'",
     tokens: (
       Identifier:"export",
@@ -2013,6 +2145,98 @@ mod tests {
     ),
   }
 
+  test! {
+    name:   format_string_empty,
+    text:   "f''",
+    tokens: (
+      Identifier: "f",
+      StringToken: "''",
+    ),
+  }
+
+  test! {
+    name:   format_string_identifier,
+    text:   "f'{{foo}}'",
+    tokens: (
+      Identifier: "f",
+      FormatStringStart: "'{{",
+      Identifier: "foo",
+      FormatStringEnd: "}}'",
+    ),
+  }
+
+  test! {
+    name:   format_string_continue,
+    text:   "f'{{foo}}bar{{baz}}'",
+    tokens: (
+      Identifier: "f",
+      FormatStringStart: "'{{",
+      Identifier: "foo",
+      FormatStringContinue: "}}bar{{",
+      Identifier: "baz",
+      FormatStringEnd: "}}'",
+    ),
+  }
+
+  test! {
+    name:   format_string_whitespace,
+    text:   "f '{{foo}}'",
+    tokens: (
+      Identifier: "f",
+      Whitespace,
+      StringToken: "'{{foo}}'",
+    ),
+  }
+
+  test! {
+    name:   format_string_wrong_identifier,
+    text:   "g'{{foo}}'",
+    tokens: (
+      Identifier: "g",
+      StringToken: "'{{foo}}'",
+    ),
+  }
+
+  test! {
+    name:   format_string_followed_by_recipe,
+    text:   "foo := f'{{'foo'}}{{'bar'}}'\nbar:",
+    tokens: (
+      Identifier: "foo",
+      Whitespace: " ",
+      ColonEquals: ":=",
+      Whitespace: " ",
+      Identifier: "f",
+      FormatStringStart: "'{{",
+      StringToken: "'foo'",
+      FormatStringContinue: "}}{{",
+      StringToken: "'bar'",
+      FormatStringEnd: "}}'",
+      Eol: "\n",
+      Identifier: "bar",
+      Colon,
+    ),
+  }
+
+  test! {
+    name:   indented_format_string_followed_by_recipe,
+    text:   "foo := f'''{{'foo'}}{{'bar'}}'''\nbar:",
+    tokens: (
+      Identifier: "foo",
+      Whitespace: " ",
+      ColonEquals: ":=",
+      Whitespace: " ",
+      Identifier: "f",
+      FormatStringStart: "'''{{",
+      StringToken: "'foo'",
+      FormatStringContinue: "}}{{",
+      StringToken: "'bar'",
+      FormatStringEnd: "}}'''",
+      Eol: "\n",
+      Identifier: "bar",
+      Colon,
+    ),
+  }
+
   error! {
     name:  tokenize_space_then_tab,
     input: "a:
@@ -2043,12 +2267,12 @@ mod tests {
 
   error! {
     name:   tokenize_unknown,
-    input:  "~",
+    input:  "%",
     offset: 0,
     line:   0,
     column: 0,
     width:  1,
-    kind:   UnknownStartOfToken,
+    kind:   UnknownStartOfToken { start: '%'},
   }
 
   error! {
@@ -2102,26 +2326,6 @@ mod tests {
     kind:   UnpairedCarriageReturn,
   }
 
-  error! {
-    name:   unknown_start_of_token_ampersand,
-    input:  " \r\n&",
-    offset: 3,
-    line:   1,
-    column: 0,
-    width:  1,
-    kind:   UnknownStartOfToken,
-  }
-
-  error! {
-    name:   unknown_start_of_token_tilde,
-    input:  "~",
-    offset: 0,
-    line:   0,
-    column: 0,
-    width:  1,
-    kind:   UnknownStartOfToken,
-  }
-
   error! {
     name:   invalid_name_start_dash,
     input:  "-foo",
@@ -2129,7 +2333,7 @@ mod tests {
     line:   0,
     column: 0,
     width:  1,
-    kind:   UnknownStartOfToken,
+    kind:   UnknownStartOfToken{ start: '-'},
   }
 
   error! {
@@ -2139,7 +2343,7 @@ mod tests {
     line:   0,
     column: 0,
     width:  1,
-    kind:   UnknownStartOfToken,
+    kind:   UnknownStartOfToken { start: '0' },
   }
 
   error! {
@@ -2203,13 +2407,13 @@ mod tests {
   }
 
   error! {
-    name:   unexpected_character_after_bang,
-    input:  "!{",
+    name:   unexpected_character_after_at,
+    input:  "@%",
     offset: 1,
     line:   0,
     column: 1,
     width:  1,
-    kind:   UnexpectedCharacter { expected: '=' },
+    kind:   UnknownStartOfToken { start: '%'},
   }
 
   error! {
@@ -2226,36 +2430,88 @@ mod tests {
     },
   }
 
+  error! {
+    name:   ampersand_eof,
+    input:  "&",
+    offset: 1,
+    line:   0,
+    column: 1,
+    width:  0,
+    kind:   UnexpectedEndOfToken {
+      expected: vec!['&'],
+    },
+  }
+
+  error! {
+    name:   ampersand_unexpected,
+    input:  "&%",
+    offset: 1,
+    line:   0,
+    column: 1,
+    width:  1,
+    kind:   UnexpectedCharacter {
+      expected: vec!['&'],
+    },
+  }
+
+  error! {
+    name:   bang_eof,
+    input:  "!",
+    offset: 1,
+    line:   0,
+    column: 1,
+    width:  0,
+    kind:   UnexpectedEndOfToken {
+      expected: vec!['=', '~'],
+    },
+  }
+
+  error! {
+    name:   unclosed_parenthesis_in_interpolation,
+    input:  "a:\n echo {{foo(}}",
+    offset:  15,
+    line:   1,
+    column: 12,
+    width:  0,
+    kind:   MismatchedClosingDelimiter {
+      close: Delimiter::Brace,
+      open: Delimiter::Paren,
+      open_line: 1,
+    },
+  }
+
   #[test]
   fn presume_error() {
+    let compile_error = Lexer::new("justfile".as_ref(), "!")
+      .presume('-')
+      .unwrap_err();
     assert_matches!(
-      Lexer::new("!").presume('-').unwrap_err(),
-      CompilationError {
-        token: Token {
-          offset: 0,
-          line:   0,
-          column: 0,
-          length: 0,
-          src:    "!",
-          kind:   Unspecified,
-        },
-        kind:  Internal {
-          message,
-        },
-      } if message == "Lexer presumed character `-`"
+      compile_error.token,
+      Token {
+        offset: 0,
+        line: 0,
+        column: 0,
+        length: 0,
+        src: "!",
+        kind: Unspecified,
+        path: _,
+      }
+    );
+    assert_matches!(&*compile_error.kind,
+        Internal { message }
+        if message == "Lexer presumed character `-`"
     );
 
     assert_eq!(
-      Lexer::new("!").presume('-').unwrap_err().to_string(),
-      unindent(
-        "
-        Internal error, this may indicate a bug in just: Lexer presumed character `-`
-        \
-         consider filing an issue: https://github.com/casey/just/issues/new
-          |
-        1 | !
-          | ^"
-      ),
+      Error::Compile { compile_error }
+        .color_display(Color::never())
+        .to_string(),
+      "error: Internal error, this may indicate a bug in just: Lexer presumed character `-`
+consider filing an issue: https://github.com/casey/just/issues/new
+ ——▶ justfile:1:1
+  │
+1 │ !
+  │ ^"
     );
   }
 }
