@@ -141,7 +141,7 @@ impl<'run, 'src> Parser<'run, 'src> {
     Err(self.internal_error("`Parser::advance()` advanced past end of token stream")?)
   }
 
-  /// Return the next token if it is of kind `expected`, otherwise, return an
+  /// Return next token if it is of kind `expected`, otherwise, return an
   /// unexpected token error
   fn expect(&mut self, expected: TokenKind) -> CompileResult<'src, Token<'src>> {
     if let Some(token) = self.accept(expected)? {
@@ -151,6 +151,18 @@ impl<'run, 'src> Parser<'run, 'src> {
     }
   }
 
+  /// Return the next token if it is any of kinds in `expected`, otherwise,
+  /// return an unexpected token error
+  fn expect_any(&mut self, expected: &[TokenKind]) -> CompileResult<'src, Token<'src>> {
+    for &kind in expected {
+      if let Some(token) = self.accept(kind)? {
+        return Ok(token);
+      }
+    }
+
+    Err(self.unexpected_token()?)
+  }
+
   /// Return an unexpected token error if the next token is not an EOL
   fn expect_eol(&mut self) -> CompileResult<'src> {
     self.accept(Comment)?;
@@ -246,35 +258,39 @@ impl<'run, 'src> Parser<'run, 'src> {
     }
   }
 
-  /// Accept a token of kind `Identifier` and parse into a `Name`
-  fn accept_name(&mut self) -> CompileResult<'src, Option<Name<'src>>> {
+  /// Accept a double-colon separated sequence of identifiers
+  fn accept_namepath(&mut self) -> CompileResult<'src, Option<Namepath<'src>>> {
     if self.next_is(Identifier) {
-      Ok(Some(self.parse_name()?))
+      Ok(Some(self.parse_namepath()?))
+    } else {
+      Ok(None)
+    }
+  }
+
+  fn accept_keyword(&mut self, keyword: Keyword) -> CompileResult<'src, Option<Name<'src>>> {
+    let next = self.next()?;
+
+    if next.kind == Identifier && next.lexeme() == keyword.lexeme() {
+      self.advance()?;
+      Ok(Some(Name::from_identifier(next)))
     } else {
       Ok(None)
     }
   }
 
   fn accepted_keyword(&mut self, keyword: Keyword) -> CompileResult<'src, bool> {
-    let next = self.next()?;
-
-    if next.kind == Identifier && next.lexeme() == keyword.lexeme() {
-      self.advance()?;
-      Ok(true)
-    } else {
-      Ok(false)
-    }
+    Ok(self.accept_keyword(keyword)?.is_some())
   }
 
   /// Accept a dependency
   fn accept_dependency(&mut self) -> CompileResult<'src, Option<UnresolvedDependency<'src>>> {
-    if let Some(recipe) = self.accept_name()? {
+    if let Some(recipe) = self.accept_namepath()? {
       Ok(Some(UnresolvedDependency {
         arguments: Vec::new(),
         recipe,
       }))
     } else if self.accepted(ParenL)? {
-      let recipe = self.parse_name()?;
+      let recipe = self.parse_namepath()?;
 
       let mut arguments = Vec::new();
 
@@ -362,11 +378,10 @@ impl<'run, 'src> Parser<'run, 'src> {
           {
             self.presume_keyword(Keyword::Import)?;
             let optional = self.accepted(QuestionMark)?;
-            let (path, relative) = self.parse_string_literal_token()?;
+            let relative = self.parse_string_literal()?;
             items.push(Item::Import {
               absolute: None,
               optional,
-              path,
               relative,
             });
           }
@@ -411,7 +426,7 @@ impl<'run, 'src> Parser<'run, 'src> {
             let mut groups = Vec::new();
             for attribute in attributes {
               if let Attribute::Group(group) = attribute {
-                groups.push(group.cooked);
+                groups.push(group);
               }
             }
 
@@ -441,9 +456,9 @@ impl<'run, 'src> Parser<'run, 'src> {
             } else {
               let doc = pop_doc_comment(&mut items, eol_since_last_comment);
               items.push(Item::Recipe(self.parse_recipe(
+                take_attributes(),
                 doc,
                 false,
-                take_attributes(),
               )?));
             }
           }
@@ -451,9 +466,9 @@ impl<'run, 'src> Parser<'run, 'src> {
       } else if self.accepted(At)? {
         let doc = pop_doc_comment(&mut items, eol_since_last_comment);
         items.push(Item::Recipe(self.parse_recipe(
+          take_attributes(),
           doc,
           true,
-          take_attributes(),
         )?));
       } else {
         return Err(self.unexpected_token()?);
@@ -475,6 +490,10 @@ impl<'run, 'src> Parser<'run, 'src> {
 
     Ok(Ast {
       items,
+      module_path: self
+        .module_namepath
+        .map(ToString::to_string)
+        .unwrap_or_default(),
       unstable_features: self.unstable_features,
       warnings: Vec::new(),
       working_directory: self.working_directory.into(),
@@ -517,10 +536,10 @@ impl<'run, 'src> Parser<'run, 'src> {
     attributes.ensure_valid_attributes("Assignment", *name, &[AttributeDiscriminant::Private])?;
 
     Ok(Assignment {
-      constant: false,
       export,
       file_depth: self.file_depth,
       name,
+      prelude: false,
       private: private || name.lexeme().starts_with('_'),
       value,
     })
@@ -645,6 +664,29 @@ impl<'run, 'src> Parser<'run, 'src> {
     })
   }
 
+  fn parse_format_string(&mut self) -> CompileResult<'src, Expression<'src>> {
+    self.expect_keyword(Keyword::F)?;
+
+    let start = self.parse_string_literal_in_state(StringState::FormatStart)?;
+
+    let kind = StringKind::from_string_or_backtick(start.token)?;
+
+    let mut more = start.token.kind == FormatStringStart;
+
+    let mut expressions = Vec::new();
+
+    while more {
+      let expression = self.parse_expression()?;
+      more = self.next_is(FormatStringContinue);
+      expressions.push((
+        expression,
+        self.parse_string_literal_in_state(StringState::FormatContinue(kind))?,
+      ));
+    }
+
+    Ok(Expression::FormatString { start, expressions })
+  }
+
   // Check if the next tokens are a shell-expanded string, i.e., `x"foo"`.
   //
   // This function skips initial whitespace tokens, but thereafter is
@@ -659,16 +701,38 @@ impl<'run, 'src> Parser<'run, 'src> {
 
     tokens
       .next()
-      .is_some_and(|token| token.kind == Identifier && token.lexeme() == "x")
+      .is_some_and(|token| token.kind == Identifier && token.lexeme() == Keyword::X.lexeme())
       && tokens.next().is_some_and(|token| token.kind == StringToken)
   }
 
+  // Check if the next tokens are a format string, i.e., `f"foo"`.
+  //
+  // This function skips initial whitespace tokens, but thereafter is
+  // whitespace-sensitive, so `f"foo"` is a format string, whereas `f
+  // "foo"` is not.
+  fn next_is_format_string(&self) -> bool {
+    let mut tokens = self
+      .tokens
+      .iter()
+      .skip(self.next_token)
+      .skip_while(|token| token.kind == Whitespace);
+
+    tokens
+      .next()
+      .is_some_and(|token| token.kind == Identifier && token.lexeme() == Keyword::F.lexeme())
+      && tokens
+        .next()
+        .is_some_and(|token| matches!(token.kind, StringToken | FormatStringStart))
+  }
+
   /// Parse a value, e.g. `(bar)`
   fn parse_value(&mut self) -> CompileResult<'src, Expression<'src>> {
     if self.next_is(StringToken) || self.next_is_shell_expanded_string() {
       Ok(Expression::StringLiteral {
         string_literal: self.parse_string_literal()?,
       })
+    } else if self.next_is_format_string() {
+      self.parse_format_string()
     } else if self.next_is(Backtick) {
       let next = self.next()?;
       let kind = StringKind::from_string_or_backtick(next)?;
@@ -686,13 +750,17 @@ impl<'run, 'src> Parser<'run, 'src> {
       }
       Ok(Expression::Backtick { contents, token })
     } else if self.next_is(Identifier) {
-      if self.accepted_keyword(Keyword::Assert)? {
+      if let Some(name) = self.accept_keyword(Keyword::Assert)? {
         self.expect(ParenL)?;
         let condition = self.parse_condition()?;
         self.expect(Comma)?;
         let error = Box::new(self.parse_expression()?);
         self.expect(ParenR)?;
-        Ok(Expression::Assert { condition, error })
+        Ok(Expression::Assert {
+          condition,
+          error,
+          name,
+        })
       } else {
         let name = self.parse_name()?;
 
@@ -720,10 +788,16 @@ impl<'run, 'src> Parser<'run, 'src> {
     }
   }
 
-  /// Parse a string literal, e.g. `"FOO"`, returning the string literal and the string token
-  fn parse_string_literal_token(
+  /// Parse a string literal, e.g. `"FOO"`
+  fn parse_string_literal(&mut self) -> CompileResult<'src, StringLiteral<'src>> {
+    self.parse_string_literal_in_state(StringState::Normal)
+  }
+
+  /// Parse a string literal, e.g. `"FOO"`
+  fn parse_string_literal_in_state(
     &mut self,
-  ) -> CompileResult<'src, (Token<'src>, StringLiteral<'src>)> {
+    state: StringState,
+  ) -> CompileResult<'src, StringLiteral<'src>> {
     let expand = if self.next_is(Identifier) {
       self.expect_keyword(Keyword::X)?;
       true
@@ -731,24 +805,49 @@ impl<'run, 'src> Parser<'run, 'src> {
       false
     };
 
-    let token = self.expect(StringToken)?;
+    let token = match state {
+      StringState::Normal => self.expect(StringToken)?,
+      StringState::FormatStart => self.expect_any(&[StringToken, FormatStringStart])?,
+      StringState::FormatContinue(_) => {
+        self.expect_any(&[FormatStringContinue, FormatStringEnd])?
+      }
+    };
 
-    let kind = StringKind::from_string_or_backtick(token)?;
+    let kind = match state {
+      StringState::Normal | StringState::FormatStart => StringKind::from_string_or_backtick(token)?,
+      StringState::FormatContinue(kind) => kind,
+    };
 
-    let delimiter_len = kind.delimiter_len();
+    let open = if matches!(token.kind, FormatStringContinue | FormatStringEnd) {
+      Lexer::INTERPOLATION_END.len()
+    } else {
+      kind.delimiter_len()
+    };
 
-    let raw = &token.lexeme()[delimiter_len..token.lexeme().len() - delimiter_len];
+    let close = if matches!(token.kind, FormatStringStart | FormatStringContinue) {
+      Lexer::INTERPOLATION_START.len()
+    } else {
+      kind.delimiter_len()
+    };
 
-    let unindented = if kind.indented() {
+    let raw = &token.lexeme()[open..token.lexeme().len() - close];
+
+    let unindented = if kind.indented() && matches!(token.kind, StringToken) {
       unindent(raw)
     } else {
       raw.to_owned()
     };
 
-    let cooked = if kind.processes_escape_sequences() {
-      Self::cook_string(token, &unindented)?
-    } else {
+    let undelimited = if matches!(state, StringState::Normal) {
       unindented
+    } else {
+      unindented.replace(Lexer::INTERPOLATION_ESCAPE, Lexer::INTERPOLATION_START)
+    };
+
+    let cooked = if kind.processes_escape_sequences() {
+      Self::cook_string(token, &undelimited)?
+    } else {
+      undelimited
     };
 
     let cooked = if expand {
@@ -759,15 +858,29 @@ impl<'run, 'src> Parser<'run, 'src> {
       cooked
     };
 
-    Ok((
+    Ok(StringLiteral {
       token,
-      StringLiteral {
-        cooked,
-        expand,
-        kind,
-        raw,
+      cooked,
+      expand,
+      kind,
+      part: match token.kind {
+        FormatStringStart => Some(FormatStringPart::Start),
+        FormatStringContinue => Some(FormatStringPart::Continue),
+        FormatStringEnd => Some(FormatStringPart::End),
+        StringToken => {
+          if matches!(state, StringState::Normal) {
+            None
+          } else {
+            Some(FormatStringPart::Single)
+          }
+        }
+        _ => {
+          return Err(token.error(CompileErrorKind::Internal {
+            message: "unexpected token kind while parsing string literal".into(),
+          }));
+        }
       },
-    ))
+    })
   }
 
   // Transform escape sequences in from string literal `token` with content `text`
@@ -805,7 +918,7 @@ impl<'run, 'src> Parser<'run, 'src> {
             '\n' => {}
             '"' => cooked.push('"'),
             character => {
-              return Err(token.error(CompileErrorKind::InvalidEscapeSequence { character }))
+              return Err(token.error(CompileErrorKind::InvalidEscapeSequence { character }));
             }
           }
           state = State::Initial;
@@ -852,12 +965,6 @@ impl<'run, 'src> Parser<'run, 'src> {
     Ok(cooked)
   }
 
-  /// Parse a string literal, e.g. `"FOO"`
-  fn parse_string_literal(&mut self) -> CompileResult<'src, StringLiteral<'src>> {
-    let (_token, string_literal) = self.parse_string_literal_token()?;
-    Ok(string_literal)
-  }
-
   /// Parse a name from an identifier token
   fn parse_name(&mut self) -> CompileResult<'src, Name<'src>> {
     self.expect(Identifier).map(Name::from_identifier)
@@ -898,16 +1005,73 @@ impl<'run, 'src> Parser<'run, 'src> {
   /// Parse a recipe
   fn parse_recipe(
     &mut self,
+    attributes: AttributeSet<'src>,
     doc: Option<&'src str>,
     quiet: bool,
-    attributes: AttributeSet<'src>,
   ) -> CompileResult<'src, UnresolvedRecipe<'src>> {
     let name = self.parse_name()?;
 
     let mut positional = Vec::new();
 
+    let mut longs = HashSet::new();
+    let mut shorts = HashSet::new();
+
+    let mut arg_attributes = BTreeMap::new();
+
+    for attribute in &attributes {
+      let Attribute::Arg {
+        help,
+        long,
+        long_key,
+        name: arg,
+        pattern,
+        short,
+        value,
+        ..
+      } = attribute
+      else {
+        continue;
+      };
+
+      if let Some(option) = long {
+        if !longs.insert(&option.cooked) {
+          return Err(
+            long_key
+              .unwrap_or(option.token)
+              .error(CompileErrorKind::DuplicateOption {
+                option: Switch::Long(option.cooked.clone()),
+                recipe: name.lexeme(),
+              }),
+          );
+        }
+      }
+
+      if let Some(option) = short {
+        if !shorts.insert(&option.cooked) {
+          return Err(option.token.error(CompileErrorKind::DuplicateOption {
+            option: Switch::Short(option.cooked.chars().next().unwrap()),
+            recipe: name.lexeme(),
+          }));
+        }
+      }
+
+      arg_attributes.insert(
+        arg.cooked.clone(),
+        ArgAttribute {
+          help: help.as_ref().map(|literal| literal.cooked.clone()),
+          name: arg.token,
+          pattern: pattern.clone(),
+          long: long.as_ref().map(|long| long.cooked.clone()),
+          short: short
+            .as_ref()
+            .map(|short| short.cooked.chars().next().unwrap()),
+          value: value.as_ref().map(|value| value.cooked.clone()),
+        },
+      );
+    }
+
     while self.next_is(Identifier) || self.next_is(Dollar) {
-      positional.push(self.parse_parameter(ParameterKind::Singular)?);
+      positional.push(self.parse_parameter(&mut arg_attributes, ParameterKind::Singular)?);
     }
 
     let kind = if self.accepted(Plus)? {
@@ -919,7 +1083,7 @@ impl<'run, 'src> Parser<'run, 'src> {
     };
 
     let variadic = if kind.is_variadic() {
-      let variadic = self.parse_parameter(kind)?;
+      let variadic = self.parse_parameter(&mut arg_attributes, kind)?;
 
       self.forbid(Identifier, |token| {
         token.error(CompileErrorKind::ParameterFollowsVariadicParameter {
@@ -934,6 +1098,10 @@ impl<'run, 'src> Parser<'run, 'src> {
 
     self.expect(Colon)?;
 
+    if let Some((argument, ArgAttribute { name, .. })) = arg_attributes.into_iter().next() {
+      return Err(name.error(CompileErrorKind::UndefinedArgAttribute { argument }));
+    }
+
     let mut dependencies = Vec::new();
 
     while let Some(dependency) = self.accept_dependency()? {
@@ -963,12 +1131,6 @@ impl<'run, 'src> Parser<'run, 'src> {
     let shebang = body.first().is_some_and(Line::is_shebang);
     let script = attributes.contains(AttributeDiscriminant::Script);
 
-    if shebang && script {
-      return Err(name.error(CompileErrorKind::ShebangAndScriptAttribute {
-        recipe: name.lexeme(),
-      }));
-    }
-
     if attributes.contains(AttributeDiscriminant::WorkingDirectory)
       && attributes.contains(AttributeDiscriminant::NoCd)
     {
@@ -1009,9 +1171,7 @@ impl<'run, 'src> Parser<'run, 'src> {
       file_depth: self.file_depth,
       import_offsets: self.import_offsets.clone(),
       name,
-      namepath: self
-        .module_namepath
-        .map_or_else(|| name.into(), |module_namepath| module_namepath.join(name)),
+      namepath: None,
       parameters: positional.into_iter().chain(variadic).collect(),
       priors,
       private,
@@ -1020,7 +1180,11 @@ impl<'run, 'src> Parser<'run, 'src> {
   }
 
   /// Parse a recipe parameter
-  fn parse_parameter(&mut self, kind: ParameterKind) -> CompileResult<'src, Parameter<'src>> {
+  fn parse_parameter(
+    &mut self,
+    arg_attributes: &mut BTreeMap<String, ArgAttribute<'src>>,
+    kind: ParameterKind,
+  ) -> CompileResult<'src, Parameter<'src>> {
     let export = self.accepted(Dollar)?;
 
     let name = self.parse_name()?;
@@ -1031,11 +1195,34 @@ impl<'run, 'src> Parser<'run, 'src> {
       None
     };
 
+    let mut help = None;
+    let mut long = None;
+    let mut pattern = None;
+    let mut short = None;
+    let mut value = None;
+
+    if let Some(arg) = arg_attributes.remove(name.lexeme()) {
+      help = arg.help;
+      long = arg.long;
+      pattern = arg.pattern;
+      short = arg.short;
+      value = arg.value;
+    }
+
+    if kind.is_variadic() && (long.is_some() || short.is_some()) {
+      return Err(name.error(CompileErrorKind::VariadicParameterWithOption));
+    }
+
     Ok(Parameter {
       default,
       export,
+      help,
       kind,
+      long,
       name,
+      pattern,
+      short,
+      value,
     })
   }
 
@@ -1139,13 +1326,13 @@ impl<'run, 'src> Parser<'run, 'src> {
     self.expect(ColonEquals)?;
 
     let set_value = match keyword {
-      Keyword::DotenvFilename => Some(Setting::DotenvFilename(self.parse_string_literal()?)),
-      Keyword::DotenvPath => Some(Setting::DotenvPath(self.parse_string_literal()?)),
+      Keyword::DotenvFilename => Some(Setting::DotenvFilename(self.parse_expression()?)),
+      Keyword::DotenvPath => Some(Setting::DotenvPath(self.parse_expression()?)),
       Keyword::ScriptInterpreter => Some(Setting::ScriptInterpreter(self.parse_interpreter()?)),
       Keyword::Shell => Some(Setting::Shell(self.parse_interpreter()?)),
-      Keyword::Tempdir => Some(Setting::Tempdir(self.parse_string_literal()?)),
+      Keyword::Tempdir => Some(Setting::Tempdir(self.parse_expression()?)),
       Keyword::WindowsShell => Some(Setting::WindowsShell(self.parse_interpreter()?)),
-      Keyword::WorkingDirectory => Some(Setting::WorkingDirectory(self.parse_string_literal()?)),
+      Keyword::WorkingDirectory => Some(Setting::WorkingDirectory(self.parse_expression()?)),
       _ => None,
     };
 
@@ -1159,16 +1346,16 @@ impl<'run, 'src> Parser<'run, 'src> {
   }
 
   /// Parse interpreter setting value, i.e., `['sh', '-eu']`
-  fn parse_interpreter(&mut self) -> CompileResult<'src, Interpreter<'src>> {
+  fn parse_interpreter(&mut self) -> CompileResult<'src, Interpreter<Expression<'src>>> {
     self.expect(BracketL)?;
 
-    let command = self.parse_string_literal()?;
+    let command = self.parse_expression()?;
 
     let mut arguments = Vec::new();
 
     if self.accepted(Comma)? {
       while !self.next_is(BracketR) {
-        arguments.push(self.parse_string_literal()?);
+        arguments.push(self.parse_expression()?);
 
         if !self.accepted(Comma)? {
           break;
@@ -1183,7 +1370,8 @@ impl<'run, 'src> Parser<'run, 'src> {
 
   /// Item attributes, i.e., `[macos]` or `[confirm: "warning!"]`
   fn parse_attributes(&mut self) -> CompileResult<'src, Option<(Token<'src>, AttributeSet<'src>)>> {
-    let mut attributes = BTreeMap::new();
+    let mut arg_attributes = BTreeMap::new();
+    let mut attributes = Vec::new();
     let mut discriminants = BTreeMap::new();
 
     let mut token = None;
@@ -1195,29 +1383,50 @@ impl<'run, 'src> Parser<'run, 'src> {
         let name = self.parse_name()?;
 
         let mut arguments = Vec::new();
+        let mut keyword_arguments = BTreeMap::new();
 
         if self.accepted(Colon)? {
           arguments.push(self.parse_string_literal()?);
         } else if self.accepted(ParenL)? {
           loop {
-            arguments.push(self.parse_string_literal()?);
+            if self.next_is(Identifier) && !self.next_is_shell_expanded_string() {
+              let key = self.parse_name()?;
 
-            if !self.accepted(Comma)? {
+              let value = self
+                .accepted(Equals)?
+                .then(|| self.parse_string_literal())
+                .transpose()?;
+
+              keyword_arguments.insert(key.lexeme(), (key, value));
+            } else {
+              let literal = self.parse_string_literal()?;
+
+              if !keyword_arguments.is_empty() {
+                return Err(
+                  literal
+                    .token
+                    .error(CompileErrorKind::AttributePositionalFollowsKeyword),
+                );
+              }
+
+              arguments.push(literal);
+            }
+
+            if !self.accepted(Comma)? || self.next_is(ParenR) {
               break;
             }
           }
+
           self.expect(ParenR)?;
         }
 
-        let attribute = Attribute::new(name, arguments)?;
+        let attribute = Attribute::new(name, arguments, keyword_arguments)?;
 
-        let first = attributes.get(&attribute).or_else(|| {
-          if attribute.repeatable() {
-            None
-          } else {
-            discriminants.get(&attribute.discriminant())
-          }
-        });
+        let first = if attribute.repeatable() {
+          None
+        } else {
+          discriminants.get(&attribute.discriminant())
+        };
 
         if let Some(&first) = first {
           return Err(name.error(CompileErrorKind::DuplicateAttribute {
@@ -1226,9 +1435,20 @@ impl<'run, 'src> Parser<'run, 'src> {
           }));
         }
 
+        if let Attribute::Arg { name: arg, .. } = &attribute {
+          if let Some(&first) = arg_attributes.get(&arg.cooked) {
+            return Err(name.error(CompileErrorKind::DuplicateArgAttribute {
+              arg: arg.cooked.clone(),
+              first,
+            }));
+          }
+
+          arg_attributes.insert(arg.cooked.clone(), name.line);
+        }
+
         discriminants.insert(attribute.discriminant(), name.line);
 
-        attributes.insert(attribute, name.line);
+        attributes.push(attribute);
 
         if !self.accepted(Comma)? {
           break;
@@ -1241,7 +1461,7 @@ impl<'run, 'src> Parser<'run, 'src> {
     if attributes.is_empty() {
       Ok(None)
     } else {
-      Ok(Some((token.unwrap(), attributes.into_keys().collect())))
+      Ok(Some((token.unwrap(), attributes.into_iter().collect())))
     }
   }
 }
@@ -1299,6 +1519,7 @@ mod tests {
     };
   }
 
+  #[track_caller]
   fn error(
     src: &str,
     offset: usize,
@@ -1631,6 +1852,24 @@ mod tests {
     tree: (justfile (recipe foo (body ("bar")))),
   }
 
+  test! {
+    name: recipe_dependency_module,
+    text: "foo: bar::baz",
+    tree: (justfile (recipe foo (deps (bar baz)))),
+  }
+
+  test! {
+    name: recipe_dependency_parenthesis_module,
+    text: "foo: (bar::baz)",
+    tree: (justfile (recipe foo (deps (bar baz)))),
+  }
+
+  test! {
+    name: recipe_dependency_module_mixed,
+      text: "foo: bar::baz qux",
+    tree: (justfile (recipe foo (deps (bar baz) qux))),
+  }
+
   test! {
     name: recipe_line_multiple,
     text: "foo:\n bar\n baz\n {{\"bob\"}}biz",
@@ -2437,6 +2676,24 @@ mod tests {
     tree: (justfile (assignment foo (assert (if a != b c d) == "abc" "error"))),
   }
 
+  test! {
+    name: format_string_simple,
+    text: "foo := f'abc'",
+    tree: (justfile (assignment foo (format "abc"))),
+  }
+
+  test! {
+    name: format_string_expression,
+    text: "foo := f'foo{{ 'abc' + 'xyz' }}bar'",
+    tree: (justfile (assignment foo (format "foo" (+ "abc" "xyz") "bar"))),
+  }
+
+  test! {
+    name: format_string_complex,
+    text: "foo := f'foo{{ 'abc' + 'xyz' }}bar{{ 'hello' }}goodbye'",
+    tree: (justfile (assignment foo (format "foo" (+ "abc" "xyz") "bar" "hello" "goodbye"))),
+  }
+
   error! {
     name:   alias_syntax_multiple_rhs,
     input:  "alias foo := bar baz",
@@ -2534,7 +2791,7 @@ mod tests {
     column:  9,
     width:   1,
     kind:    UnexpectedToken{
-      expected: vec![AmpersandAmpersand, Comment, Eof, Eol, Identifier, ParenL],
+      expected: vec![AmpersandAmpersand, ColonColon, Comment, Eof, Eol, Identifier, ParenL],
       found: Equals
     },
   }
@@ -2572,26 +2829,6 @@ mod tests {
     },
   }
 
-  error! {
-    name:   unclosed_parenthesis_in_interpolation,
-    input:  "a:\n echo {{foo(}}",
-    offset:  15,
-    line:   1,
-    column: 12,
-    width:  2,
-    kind:   UnexpectedToken{
-      expected: vec![
-        Backtick,
-        Identifier,
-        ParenL,
-        ParenR,
-        Slash,
-        StringToken,
-      ],
-      found: InterpolationEnd,
-    },
-  }
-
   error! {
     name:   plus_following_parameter,
     input:  "a b c+:",
@@ -2664,39 +2901,16 @@ mod tests {
     width:  1,
     kind:   UnexpectedToken {
       expected: vec![
+        Backtick,
         Identifier,
+        ParenL,
+        Slash,
         StringToken,
       ],
       found: BracketR,
     },
   }
 
-  error! {
-    name:   set_shell_non_literal_first,
-    input:  "set shell := ['bar' + 'baz']",
-    offset: 20,
-    line:   0,
-    column: 20,
-    width:  1,
-    kind:   UnexpectedToken {
-      expected: vec![BracketR, Comma],
-      found: Plus,
-    },
-  }
-
-  error! {
-    name:   set_shell_non_literal_second,
-    input:  "set shell := ['biz', 'bar' + 'baz']",
-    offset: 27,
-    line:   0,
-    column: 27,
-    width:  1,
-    kind:   UnexpectedToken {
-      expected: vec![BracketR, Comma],
-      found: Plus,
-    },
-  }
-
   error! {
     name:   set_shell_bad_comma,
     input:  "set shell := ['bash',",
@@ -2706,8 +2920,11 @@ mod tests {
     width:  0,
     kind:   UnexpectedToken {
       expected: vec![
+        Backtick,
         BracketR,
         Identifier,
+        ParenL,
+        Slash,
         StringToken,
       ],
       found: Eof,
@@ -2722,7 +2939,7 @@ mod tests {
     column: 20,
     width:  0,
     kind:   UnexpectedToken {
-      expected: vec![BracketR, Comma],
+      expected: vec![AmpersandAmpersand, BarBar, BracketR, Comma, Plus, Slash],
       found: Eof,
     },
   }
