@@ -14,7 +14,7 @@ struct Assignment<'a> {
   export: bool,
   name: &'a str,
   private: bool,
-  value: &'a str,
+  value: serde_json::Value,
 }
 
 #[derive(Debug, Default, Deserialize, PartialEq, Serialize)]
@@ -52,8 +52,13 @@ struct Module<'a> {
 struct Parameter<'a> {
   default: Option<&'a str>,
   export: bool,
+  help: Option<&'a str>,
   kind: &'a str,
+  long: Option<&'a str>,
   name: &'a str,
+  pattern: Option<&'a str>,
+  short: Option<char>,
+  value: Option<&'a str>,
 }
 
 #[derive(Debug, Default, Deserialize, PartialEq, Serialize)]
@@ -180,7 +185,7 @@ fn assignment() {
         "foo",
         Assignment {
           name: "foo",
-          value: "bar",
+          value: "bar".into(),
           ..default()
         },
       )]
@@ -204,7 +209,7 @@ fn private_assignment() {
           "_foo",
           Assignment {
             name: "_foo",
-            value: "foo",
+            value: "foo".into(),
             private: true,
             ..default()
           },
@@ -213,7 +218,7 @@ fn private_assignment() {
           "bar",
           Assignment {
             name: "bar",
-            value: "bar",
+            value: "bar".into(),
             private: true,
             ..default()
           },
@@ -314,7 +319,7 @@ fn dependency_argument() {
         "x",
         Assignment {
           name: "x",
-          value: "foo",
+          value: "foo".into(),
           ..default()
         },
       )]
@@ -426,7 +431,7 @@ fn duplicate_variables() {
         "x",
         Assignment {
           name: "x",
-          value: "bar",
+          value: "bar".into(),
           ..default()
         },
       )]
@@ -981,3 +986,256 @@ fn doc_attribute_overrides_comment() {
     },
   );
 }
+
+#[test]
+fn format_string() {
+  case(
+    "
+      foo := f'abc'
+    ",
+    Module {
+      assignments: [(
+        "foo",
+        Assignment {
+          name: "foo",
+          value: json!(["format", "abc"]),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+  case(
+    "
+      foo := f'abc{{'bar'}}xyz'
+    ",
+    Module {
+      assignments: [(
+        "foo",
+        Assignment {
+          name: "foo",
+          value: json!(["format", "abc", "bar", "xyz"]),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+  case(
+    "
+      foo := f'abc{{'bar'}}xyz{{'baz' + 'buzz'}}123'
+    ",
+    Module {
+      assignments: [(
+        "foo",
+        Assignment {
+          name: "foo",
+          value: json!([
+            "format",
+            "abc",
+            "bar",
+            "xyz",
+            ["concatenate", "baz", "buzz"],
+            "123"
+          ]),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+}
+
+#[test]
+fn arg_pattern() {
+  case(
+    "[arg('bar', pattern='BAR')]\nfoo bar:",
+    Module {
+      first: Some("foo"),
+      recipes: [(
+        "foo",
+        Recipe {
+          name: "foo",
+          namepath: "foo",
+          attributes: [json!({
+            "arg": {
+              "help": null,
+              "long": null,
+              "name": "bar",
+              "pattern": "BAR",
+              "short": null,
+              "value": null,
+            }
+          })]
+          .into(),
+          parameters: [Parameter {
+            kind: "singular",
+            name: "bar",
+            pattern: Some("BAR"),
+            ..default()
+          }]
+          .into(),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+}
+
+#[test]
+fn arg_long() {
+  case(
+    "[arg('bar', long='BAR')]\nfoo bar:",
+    Module {
+      first: Some("foo"),
+      recipes: [(
+        "foo",
+        Recipe {
+          name: "foo",
+          namepath: "foo",
+          attributes: [json!({
+            "arg": {
+              "help": null,
+              "long": "BAR",
+              "name": "bar",
+              "pattern": null,
+              "short": null,
+              "value": null,
+            }
+          })]
+          .into(),
+          parameters: [Parameter {
+            kind: "singular",
+            name: "bar",
+            long: Some("BAR"),
+            ..default()
+          }]
+          .into(),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+}
+
+#[test]
+fn arg_short() {
+  case(
+    "[arg('bar', short='B')]\nfoo bar:",
+    Module {
+      first: Some("foo"),
+      recipes: [(
+        "foo",
+        Recipe {
+          name: "foo",
+          namepath: "foo",
+          attributes: [json!({
+            "arg": {
+              "help": null,
+              "long": null,
+              "name": "bar",
+              "pattern": null,
+              "short": "B",
+              "value": null,
+            }
+          })]
+          .into(),
+          parameters: [Parameter {
+            kind: "singular",
+            name: "bar",
+            short: Some('B'),
+            ..default()
+          }]
+          .into(),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+}
+
+#[test]
+fn arg_value() {
+  case(
+    "[arg('bar', short='B', value='hello')]\nfoo bar:",
+    Module {
+      first: Some("foo"),
+      recipes: [(
+        "foo",
+        Recipe {
+          name: "foo",
+          namepath: "foo",
+          attributes: [json!({
+            "arg": {
+              "help": null,
+              "long": null,
+              "name": "bar",
+              "pattern": null,
+              "short": "B",
+              "value": "hello",
+            }
+          })]
+          .into(),
+          parameters: [Parameter {
+            kind: "singular",
+            name: "bar",
+            short: Some('B'),
+            value: Some("hello"),
+            ..default()
+          }]
+          .into(),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+}
+
+#[test]
+fn arg_help() {
+  case(
+    "[arg('bar', help='hello')]\nfoo bar:",
+    Module {
+      first: Some("foo"),
+      recipes: [(
+        "foo",
+        Recipe {
+          name: "foo",
+          namepath: "foo",
+          attributes: [json!({
+            "arg": {
+              "help": "hello",
+              "long": null,
+              "name": "bar",
+              "pattern": null,
+              "short": null,
+              "value": null,
+            }
+          })]
+          .into(),
+          parameters: [Parameter {
+            help: Some("hello"),
+            kind: "singular",
+            name: "bar",
+            ..default()
+          }]
+          .into(),
+          ..default()
+        },
+      )]
+      .into(),
+      ..default()
+    },
+  );
+}
